<?php

namespace Drupal\charts_apexcharts\Plugin\chart\Library;

use Drupal\charts\Element\Chart as ChartElement;
use Drupal\charts\Plugin\chart\Library\ChartBase;
use Drupal\charts\TypeManager;
use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Form\FormBuilder;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Render\Element;
use Drupal\Core\Render\ElementInfoManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Define a concrete class for a Chart.
 *
 * @Chart(
 *   id = "apexcharts",
 *   name = @Translation("Apexcharts"),
 *   types = {
 *     "area",
 *     "bar",
 *     "boxplot",
 *     "bubble",
 *     "candlestick",
 *     "column",
 *     "donut",
 *     "gauge",
 *     "line",
 *     "pie",
 *     "radar",
 *     "scatter",
 *     "spline",
 *   },
 * )
 */
class Apexcharts extends ChartBase implements ContainerFactoryPluginInterface {

  /**
   * The element info manager.
   *
   * @var \Drupal\Core\Render\ElementInfoManagerInterface
   */
  protected $elementInfo;

  /**
   * The chart type manager.
   *
   * @var \Drupal\charts\TypeManager
   */
  protected $chartTypeManager;

  /**
   * The chart type manager.
   *
   * @var \Drupal\Core\Form\FormBuilder
   */
  protected $formBuilder;

  /**
   * Constructs a \Drupal\views\Plugin\Block\ViewsBlockBase object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Render\ElementInfoManagerInterface $element_info
   *   The element info manager.
   * @param \Drupal\charts\TypeManager $chart_type_manager
   *   The chart type manager.
   * @param \Drupal\Core\Form\FormBuilder $form_builder
   *   The form builder.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, ElementInfoManagerInterface $element_info, TypeManager $chart_type_manager, FormBuilder $form_builder) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->elementInfo = $element_info;
    $this->chartTypeManager = $chart_type_manager;
    $this->formBuilder = $form_builder;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('element_info'),
      $container->get('plugin.manager.charts_type'),
      $container->get('form_builder'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function preRender(array $element) {
    $chart_definition = [];

    if (!isset($element['#id'])) {
      $element['#id'] = Html::getUniqueId('apexcharts-render');
    }
    $chart_definition = $this->populateOptions($element, $chart_definition);
    $chart_definition = $this->populateAxes($element, $chart_definition);
    $chart_definition = $this->populateData($element, $chart_definition);

    $element['#attached']['library'][] = 'charts_apexcharts/apexcharts';
    $element['#attributes']['class'][] = 'charts-apexcharts';
    $element['#chart_definition'] = $chart_definition;

    return $element;
  }

  /**
   * Populate options.
   *
   * @param array $element
   *   The element.
   * @param array $chart_definition
   *   The chart definition.
   *
   * @return array
   *   Return the chart definition.
   */
  protected function populateOptions(array $element, array $chart_definition) {
    $chart_type = $this->chartTypeConversion($element['#chart_type']);
    if ($element['#chart_type'] === 'bar') {
      $chart_definition['plotOptions']['bar']['horizontal'] = TRUE;
    }
    if ($chart_type === 'line' && !empty($element['#polar'])) {
      $chart_type = 'radar';
    }
    if ($element['#chart_type'] === 'spline') {
      $chart_definition['stroke']['curve'] = 'smooth';
    }
    $chart_definition['chart']['stacked'] = (bool) $element['#stacking'];

    $chart_definition['chart']['type'] = $chart_type;
    $chart_definition['title']['text'] = $element['#title'] ?? '';
    $chart_definition['subtitle']['text'] = $element['#subtitle'] ?? '';

    if ($element['#chart_type'] === 'gauge') {
      $chart_definition['fill']['gradient']['opacityFrom'] = (int) $element['#gauge']['min'];
      $chart_definition['fill']['gradient']['opacityTo'] = (int) $element['#gauge']['max'];
      $chart_definition['fill']['gradient']['type'] = 'horizontal';
      $stops = [];
      // If the following have a value, place them in the stops array.
      if ((int) $element['#gauge']['min'] === 0 || !empty($element['#gauge']['min'])) {
        $stops[] = (int) $element['#gauge']['min'];
      }
      if (!empty($element['#gauge']['yellow_from'])) {
        $stops[] = (int) $element['#gauge']['yellow_from'];
      }
      if (!empty($element['#gauge']['green_from'])) {
        $stops[] = (int) $element['#gauge']['green_from'];
      }
      if (!empty($element['#gauge']['max'])) {
        $stops[] = (int) $element['#gauge']['max'];
      }
      $chart_definition['fill']['gradient']['stops'] = $stops;
    }

    // Merge in chart raw options.
    if (!empty($element['#raw_options'])) {
      $chart_definition = NestedArray::mergeDeepArray([
        $chart_definition,
        $element['#raw_options'],
      ]);
    }

    return $chart_definition;
  }

  /**
   * Utility to populate data.
   *
   * @param array $element
   *   The element.
   * @param array $chart_definition
   *   The chart definition.
   *
   * @return array
   *   Return the chart definition.
   */
  protected function populateData(array &$element, array $chart_definition) {
    $categories = [];
    $chart_type = $this->chartTypeConversion($element['#chart_type']);
    foreach (Element::children($element) as $key) {
      if ($element[$key]['#type'] === 'chart_xaxis' && !empty($element[$key]['#labels'])) {
        if (in_array($chart_type, ['pie', 'donut'])) {
          $categories = $element[$key]['#labels'];
          break;
        }
        $categories[] = $element[$key]['#labels'];
      }
    }
    foreach (Element::children($element) as $key) {
      if ($element[$key]['#type'] === 'chart_data') {
        $series = [];
        $series_data = [];

        // Make sure defaults are loaded.
        if (empty($element[$key]['#defaults_loaded'])) {
          $element[$key] += $this->elementInfo->getInfo($element[$key]['#type']);
        }

        // Convert target named axis keys to integers.
        if (isset($element[$key]['#target_axis'])) {
          $axis_name = $element[$key]['#target_axis'];
          $axis_index = 0;
          foreach (Element::children($element) as $axis_key) {
            if ($element[$axis_key]['#type'] === 'chart_yaxis') {
              if ($axis_key === $axis_name) {
                break;
              }
              $axis_index++;
            }
          }
          $series['yaxis'] = $axis_index;
        }

        // Allow data to provide the labels.
        // This will override the axis settings.
        if ($element[$key]['#labels'] && !in_array($element[$key]['#chart_type'], [
          'scatter',
          'bubble',
        ])) {
          foreach ($element[$key]['#labels'] as $label_index => $label) {
            $series_data[$label_index][0] = $label;
          }
        }
        elseif (!empty($categories) && in_array($chart_type, ['pie', 'donut'])) {
          foreach ($categories as $label_index => $label) {
            $series_data[$label_index][0] = $label;
          }
        }

        // Populate the data.
        foreach ($element[$key]['#data'] as $data_index => $data) {
          if (isset($series_data[$data_index])) {
            $series_data[$data_index][] = $data;
          }
          elseif (in_array($chart_type, ['pie', 'donut'])) {
            $series_data[$data_index] = $data;
            $name = $series_data[$data_index]['name'] ?? NULL;
            if (!empty($name)) {
              $chart_definition['labels'][] = $name;
            }
            if (!empty($element[$key]['#grouping_colors'][$data_index][$name])) {
              $series_data[$data_index]['color'] = $element[$key]['#grouping_colors'][$data_index][$name];
            }
            elseif (!empty($element[$key]['#grouping_colors'][$data_index]) && is_array($element[$key]['#grouping_colors'][$data_index])) {
              $chart_definition['colors'][$data_index] = reset($element[$key]['#grouping_colors'][$data_index]);
            }
          }
          else {
            $series_data[$data_index] = $data;
          }
        }

        $series['type'] = !empty($element[$key]['#chart_type']) ? $element[$key]['#chart_type'] : $element['#chart_type'];
        $series['type'] = $this->chartTypeConversion($series['type']);
        if ($element['#chart_type'] === 'donut') {
          // Add innerSize to differentiate between donut and pie.
          $series['innerSize'] = '40%';
        }
        $series['name'] = $element[$key]['#title'];
        $series['color'] = $element[$key]['#color'];
        if ($series['type'] == 'line') {
          $chart_definition['stroke']['width'][] = 4;
        }
        else {
          $chart_definition['stroke']['width'][] = 0;
        }
        if ($series['type'] === 'line' && !empty($element['#polar'])) {
          unset($series['type']);
          unset($chart_definition['stroke']);
        }

        // Remove unnecessary keys to trim down the resulting JS settings.
        ChartElement::trimArray($series);

        // If you want a different type of scatter.
        if (!empty($element['#alternative_scatter'])) {
          $series = $series_data;
        }
        else {
          $series['data'] = $series_data;
        }

        // Merge in series raw options.
        if (!empty($element[$key]['#raw_options'])) {
          $series = NestedArray::mergeDeepArray([
            $series,
            $element[$key]['#raw_options'],
          ]);
        }

        // Add the series to the main chart definition.
        // Scatter colors adjustment.
        if (!empty($element['#alternative_scatter'])) {
          $chart_definition['series'] = $series;
        }
        else {
          if (in_array($series['type'], ['pie', 'donut', 'gauge', 'radialBar'])) {
            // Check if $series['data'] is in the format [1, 2, 3].
            if (is_array($series['data'][0])) {
              $series_data = array_column($series['data'], 1);
            }
            else {
              // Remove all the unnecessary keys.
              $series_data = $series['data'];
            }
            if (in_array($series['type'], ['gauge', 'radialBar'])) {
              if ($element[$key]['#color']) {
                $chart_definition['fill']['colors'] = $element[$key]['#color'];
              }
              if ($series_data[0] > 100) {
                if (!empty($element['#gauge']['max'])) {
                  $percent = round($series_data[0] / (int) $element['#gauge']['max'], 20) * 100;
                  $series_data[0] = $percent;
                }
              }
            }
            $chart_definition['series'] = $series_data;
            if (empty(array_column($series['data'], 0))) {
              if (!empty($series['name'])) {
                $chart_definition['labels'][] = $series['name'];
              }
            }
            else {
              $chart_definition['labels'] = array_column($series['data'], 0);
            }
          }
          else {
            $chart_definition['series'][] = $series;
          }
        }

        // Merge in any point-specific data points.
        foreach (Element::children($element[$key]) as $sub_key) {
          if ($element[$key][$sub_key]['#type'] === 'chart_data_item') {
            // Make sure defaults are loaded.
            if (empty($element[$key][$sub_key]['#defaults_loaded'])) {
              $element[$key][$sub_key] += $this->elementInfo->getInfo($element[$key][$sub_key]['#type']);
            }

            $data_item = $element[$key][$sub_key];
            $series_point = &$chart_definition['series'][$key]['data'][$sub_key];

            // Convert the point from a simple data value to a complex point.
            if (!isset($series_point['data'])) {
              $data = $series_point;
              $series_point = [];
              if (is_array($data)) {
                $series_point['name'] = $data[0];
                $series_point['y'] = $data[1];
              }
              else {
                $series_point['y'] = $data;
              }
            }
            if (isset($data_item['#data'])) {
              if (is_array($data_item['#data'])) {
                $series_point['x'] = $data_item['#data'][0];
                $series_point['y'] = $data_item['#data'][1];
              }
              else {
                $series_point['y'] = $data_item['#data'];
              }
            }
            if ($data_item['#title']) {
              $series_point['name'] = $data_item['#title'];
            }

            // Setting the color requires several properties for consistency.
            $series_point['color'] = $data_item['#color'];
            $series_point['fillColor'] = $data_item['#color'];
            $series_point['states']['hover']['fillColor'] = $data_item['#color'];
            $series_point['states']['select']['fillColor'] = $data_item['#color'];
            ChartElement::trimArray($series_point);

            // Merge in point raw options.
            if (!empty($data_item['#raw_options'])) {
              $series_point = NestedArray::mergeDeepArray([
                $series_point,
                $data_item['#raw_options'],
              ]);
            }
          }
        }

      }
    }

    return $chart_definition;
  }

  /**
   * Populate axes.
   *
   * @param array $element
   *   The element.
   * @param array $chart_definition
   *   The chart definition.
   *
   * @return array
   *   Return the chart definition.
   */
  protected function populateAxes(array $element, array $chart_definition) {
    foreach (Element::children($element) as $key) {
      if ($element[$key]['#type'] === 'chart_xaxis' || $element[$key]['#type'] === 'chart_yaxis') {
        // Make sure defaults are loaded.
        if (empty($element[$key]['#defaults_loaded'])) {
          $element[$key] += $this->elementInfo->getInfo($element[$key]['#type']);
        }

        // Populate the chart data.
        $axis_type = $element[$key]['#type'] === 'chart_xaxis' ? 'xaxis' : 'yaxis';
        $axis = [];
        $axis['title']['text'] = $element[$key]['#title'];
        if (!empty($element[$key]['#labels']) && !in_array($element['#chart_type'], ['pie', 'donut'])) {
          $axis['categories'] = $element[$key]['#labels'];
        }
        $axis['opposite'] = $element[$key]['#opposite'];

        // Merge in axis raw options.
        if (!empty($element[$key]['#raw_options'])) {
          $axis = NestedArray::mergeDeepArray([
            $axis,
            $element[$key]['#raw_options'],
          ]);
        }

        if ($axis_type === 'yaxis') {
          $chart_definition[$axis_type][] = $axis;
        }
        else {
          $chart_definition[$axis_type] = $axis;
        }
      }
    }

    return $chart_definition;
  }

  /**
   * Convert module chart type conventions to ApexCharts conventions.
   *
   * @param string $chart_type
   *   The chart type.
   *
   * @return string
   *   The converted chart type.
   */
  protected function chartTypeConversion(string $chart_type): string {
    $chart_type_map = [
      'column' => 'bar',
      'gauge' => 'radialBar',
      'spline' => 'line',
    ];

    return !empty($chart_type_map[$chart_type]) ? $chart_type_map[$chart_type] : $chart_type;
  }

}
